# file      : build/bootstrap.make
# author    : Boris Kolpackov <boris@kolpackov.net>
# copyright : Copyright (c) 2004 Boris Kolpackov
# license   : GNU GPL v2; http://www.gnu.org/licenses/gpl.txt




# First time bootstrap.make is read
#
ifndef %makefile%

# Configure make.
#

# Forces make to delete targets whos rebuild commands failed but
# updated the target.
#
.DELETE_ON_ERROR:

# Remove all default rules.
#
.SUFFIXES:

ifeq ($(filter -r,$(MAKEFLAGS)),)
MAKEFLAGS += -r
endif


# Disable implicit phony misfeature.
#
ifeq ($(filter --no-implicit-phony,$(MAKEFLAGS)),)
MAKEFLAGS += --no-implicit-phony
endif

# Force target. Used to force rule executions.
#
.PHONY: FORCE


# Export variables that could be useful for configuration scripts.
#
export bld_root
export MAKE

%replica%     :=
%interactive% := true

# List of realpath'ed include-once files.
#
%included_files% :=

#$(warning init %included_files%: $(%included_files%))

# @@ maybe I should set makefile to bootstrap and restore it at the end.
#    This way I will be able to use %makefile% below (instead of _self) and
#    also it will be consistent.
#

%makefile%          := $(abspath $(firstword $(MAKEFILE_LIST)))
%makefile_realpath% := $(realpath $(%makefile%))


# include facility
#
#
define include-body
%interactive%       :=
%makefile%          := $(abspath $2)
%makefile_realpath% := $(realpath $2)

$1 $2

%makefile_realpath% := $3
%makefile%          := $4
%interactive%       := $5
endef

define include
$(eval $(foreach f,$1,\
  $(call include-body,\
    include,$f,$(%makefile_realpath%),$(%makefile%),$(%interactive%))))
endef

define -include
$(eval $(foreach f,$1,\
  $(call include-body,\
    -include,$f,$(%makefile_realpath%),$(%makefile%),$(%interactive%))))
endef

#
# $1 : include/-include
# $2 : file
#

define include-once-file-body
%interactive%       :=
%makefile%          := $(abspath $2)
%makefile_realpath% := $(realpath $2)

%included_files%    += $$(%makefile_realpath%)

$1 $2

%makefile_realpath% := $3
%makefile%          := $4
%interactive%       := $5
endef

define include-once-file-filter
$(if $(filter $(realpath $2),$(%included_files%)),,\
  $(call include-once-file-body,\
    $1,$2,$(%makefile_realpath%),$(%makefile%),$(%interactive%)))
endef

#
# $1 : include/-include
# $2 : file
# $6 : value
#

define include-once-value-body
%interactive%       :=
%makefile%          := $(abspath $2)
%makefile_realpath% := $(realpath $2)

$$(if $$(%makefile_realpath%),\
  $$(eval %include_once_$$(%makefile_realpath%)% += $6))

$1 $2

%makefile_realpath% := $3
%makefile%          := $4
%interactive%       := $5
endef


define include-once-value-filter
$(if $(filter $3,$(value %include_once_$(realpath $2)%)),,\
  $(call include-once-value-body,\
    $1,$2,$(%makefile_realpath%),$(%makefile%),$(%interactive%),$3))
endef


define include-once
$(eval $(if $2,$(call include-once-value-filter,include,$1,$(strip $2)),\
               $(call include-once-file-filter,include,$1)))
endef

define -include-once
$(eval $(if $2,$(call include-once-value-filter,-include,$1,$(strip $2)),\
               $(call include-once-file-filter,-include,$1)))
endef


#$(warning include facility is up)

endif

bld_root := $(abspath $(dir $(lastword $(MAKEFILE_LIST))))

#$(warning namespace facility is up)

# frame facility
#
$(call include-once,$(bld_root)/frame.make)

# import facility
#
$(call include-once,$(bld_root)/import.make)

# load facility
#
$(call include-once,$(bld_root)/load.make)


# initialize {src,out}_{root,base}
#
define find-bootstrap-base
$(if $(findstring $(abspath $(lastword $1)),$(%makefile%)),$2,\
$(call find-bootstrap-base,\
$(wordlist 1,$(words $(wordlist 2,$(words $1),$1)),$1),\
$(lastword $1)))
endef

src_root := $(abspath $(dir $(call find-bootstrap-base,$(MAKEFILE_LIST)))..)
src_base := $(abspath $(dir $(%makefile%)))

ifneq ($(src_root),$(src_base))

ifeq ($(origin out_root),undefined)
  out_root := $(abspath $(patsubst \
    %$(subst $(src_root)/,,$(src_base)),%,$(CURDIR)))
endif

out_base := $(out_root)/$(subst $(src_root)/,,$(src_base))

else

ifeq ($(origin out_root),undefined)
  out_root := $(abspath $(CURDIR))
endif

out_base := $(out_root)

endif

scf_root := $(src_root)/build
dcf_root := $(out_root)/build

$(out_root)/%: export out_root := $(out_root)
$(out_root)/%: export src_root := $(src_root)
$(out_root)/%: export scf_root := $(scf_root)
$(out_root)/%: export dcf_root := $(dcf_root)
$(out_root)/%: export project_name := $(project_name)

$(out_base)/%: out_root := $(out_root)
$(out_base)/%: out_base := $(out_base)
$(out_base)/%: src_base := $(src_base)

ifdef build_debug

$(warning "bld root: $(bld_root)")
$(warning "scf root: $(scf_root)")
$(warning "dcf root: $(dcf_root)")
$(warning "src root: $(src_root)")
$(warning "src base: $(src_base)")
$(warning "out root: $(out_root)")
$(warning "out base: $(out_base)")

endif

#$(warning paths are set)

# Load some common facilities.
#
$(call include-once,$(bld_root)/message.make)
$(call include-once,$(bld_root)/dir.make)


# Static configuration.
#
$(call include,$(bld_root)/configuration-static.make)

ifneq ($(bld_root),$(scf_root))
$(call -include,$(scf_root)/configuration-static.make)
endif


# `disfigure' target.
#
.PHONY: disfigure
disfigure:: $(build_absolute_clean_target)

ifeq ($(.DEFAULT_GOAL),disfigure)
.DEFAULT_GOAL :=
endif


# Dynamic configuration.
#
$(call include-once,$(bld_root)/configuration-rules.make,$(dcf_root))

$(call -include,$(dcf_root)/configuration-dynamic.make)

ifndef build_configuration

.NOTPARALLEL:

endif
